SystemOrganization addCategory: #'Signals-Tests'!
SystemOrganization addCategory: #'Signals-Core'!

TestCase subclass: #SignalTest
	instanceVariableNames: 'producer consumer'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Signals-Tests'!

!SignalTest methodsFor: 'running' stamp: 'mt 6/2/2010 22:42'!
setUp

	consumer := SignalConsumer new.
	producer := SignalProducer new.! !

!SignalTest methodsFor: 'tests' stamp: 'mt 6/2/2010 22:00'!
test01ConnectDisconnect

	self assert: producer signalConnections isEmpty.
	consumer connect: producer signal: #valueChanged: toSelector: #value:.
	self assert: 1 equals: producer signalConnections size.
	
	self
		assert: (producer disconnectSignal: #valueChanged:);
		assert: producer signalConnections isEmpty;
		deny: (producer disconnectSignal: #valueChanged:).
		
	self connect: producer signal: #valueChanged: to: consumer selector: #value:.
	self assert: 1 equals: producer signalConnections size.

	self
		assert: (producer disconnectSignal: #valueChanged: from: consumer);
		assert: producer signalConnections isEmpty;
		deny: (producer disconnectSignal: #valueChanged: from: consumer).
		
	self connect: producer signal: #valueChanged: to: consumer selector: #value:.
	
	self
		assert: (producer disconnectSignal: #valueChanged: from: consumer selector: #value:);
		assert: producer signalConnections isEmpty;
		deny: (producer disconnectSignal: #valueChanged: from: consumer selector: #value:).
		
	self connect: producer signal: #valueChanged: to: consumer selector: #value:.
	self 
		deny: (producer disconnectSignal: #shouldNotExist);
		deny: producer signalConnections isEmpty.! !

!SignalTest methodsFor: 'tests' stamp: 'mt 6/2/2010 22:00'!
test02ConnectError

	self
		should: [consumer connect: producer signal: #shouldNotExist toSelector: #toggleSwitch] raise: Error;
		should: [consumer connect: producer signal: #valueChanged: toSelector: #shouldNotExist] raise: Error;
		should: [self connect: producer signal: #shouldNotExist to: consumer selector: #toggleSwitch] raise: Error;
		should: [self connect: producer signal: #valueChanged: to: consumer selector: #shouldNotExist] raise: Error.! !

!SignalTest methodsFor: 'tests' stamp: 'mt 6/2/2010 22:20'!
test03Signaling

	consumer connect: producer signal: #valueChanged: toSelector: #value:.
	consumer connect: producer signal: #newSignalProduced toSelector: #toggleSwitch.
	
	self
		assert: consumer value isNil;
		assert: #off equals: consumer switchState.	
	producer valueChanged: 'foobar'.
	producer newSignalProduced.
	self
		assert: consumer value = 'foobar';
		assert: #on equals: consumer switchState.	
	
	producer disconnectAllSignals.
	producer valueChanged: 'blubb'.
	self
		assert: consumer value = 'foobar';
		assert: #on equals: consumer switchState.! !

!SignalTest methodsFor: 'tests' stamp: 'mt 6/2/2010 22:17'!
test04Patterns

	consumer connect: producer signal: #valueChanged: toSelector: #toggleSwitch pattern: #().
	
	self assert:#off equals: consumer switchState.	
	producer valueChanged: 'foobar'.
	self assert:#on equals: consumer switchState.	! !

!SignalTest methodsFor: 'tests' stamp: 'mt 6/2/2010 22:21'!
test05PatternsReorder

	consumer connect: producer signal: #valueChanged:and: toSelector: #value: pattern: #(2).
	
	self assert: consumer value isNil.
	producer valueChanged: 'foobar' and: 'blubb'.
	self assert: 'blubb' equals: consumer value.	! !

!SignalTest methodsFor: 'tests' stamp: 'mt 6/2/2010 22:22'!
test06DoubleConnection

	consumer connect: producer signal: #valueChanged: toSelector: #value:.
	consumer connect: producer signal: #valueChanged: toSelector: #toggleSwitch pattern: #().
	
	self
		assert: consumer value isNil;
		assert: #off equals: consumer switchState.	
	
	producer valueChanged: 'foobar'.
	
	self
		assert: consumer value = 'foobar';
		assert: #on equals: consumer switchState.	! !

!SignalTest methodsFor: 'tests' stamp: 'mt 6/2/2010 22:37'!
test06PatternError

	self
		should: [consumer connect: producer signal: #valueChanged: toSelector: #toggleSwitch pattern: #(1)] raise: Error;
		should: [consumer connect: producer signal: #valueChanged: toSelector: #value: pattern: #(2)] raise: Error;
		should: [consumer connect: producer signal: #valueChanged:with: toSelector: #value: pattern: #(2 1)] raise: Error.! !

!Object methodsFor: '*signals-core' stamp: 'mt 6/2/2010 21:34'!
connect: aSender signal: aSignal to: aReceiver selector: aSelector
	"Mediator: Connect a signaling object with a receiver."

	aSender
		connectSignal: aSignal
		to: aReceiver
		selector: aSelector! !

!Object methodsFor: '*signals-core' stamp: 'mt 6/2/2010 22:09'!
connect: aSender signal: aSignal to: aReceiver selector: aSelector pattern: argPattern
	"Shadow or reorder arguments before invoking selector on receiver, e.g.,  #(1 4 3)"

	aSender
		connectSignal: aSignal
		to: aReceiver
		selector: aSelector
		pattern: argPattern! !

!Object methodsFor: '*signals-core' stamp: 'mt 6/2/2010 21:34'!
connect: aSender signal: aSignal toSelector: aSelector
	"Connect a selector from me with the signal of a sender."

	aSender
		connectSignal: aSignal
		to: self
		selector: aSelector.! !

!Object methodsFor: '*signals-core' stamp: 'mt 6/2/2010 22:09'!
connect: aSender signal: aSignal toSelector: aSelector pattern: argPattern
	"Connect a selector from me with the signal of a sender."

	aSender
		connectSignal: aSignal
		to: self
		selector: aSelector
		pattern: argPattern.! !

!Object methodsFor: '*signals-core-private' stamp: 'mt 6/2/2010 22:04'!
connectSignal: aSignal to: aReceiver selector: aSelector
	"Private: Store an incoming signal connection request."

	self
		connectSignal: aSignal
		to: aReceiver
		selector: aSelector
		pattern: nil.! !

!Object methodsFor: '*signals-core-private' stamp: 'mt 6/2/2010 22:33'!
connectSignal: aSignal to: aReceiver selector: aSelector pattern: argPattern
	"Private: Store an incoming signal connection request."

	((self respondsTo: aSignal)
		and: [aReceiver respondsTo: aSelector])
		ifFalse: [Error signal: 'Could not connect sender to receiver!!'].
	
	(argPattern notNil and: [argPattern notEmpty]) ifTrue: [
		(argPattern min <= 0
			or: [(aReceiver class >> aSelector) numArgs ~= argPattern size]
			or: [(self class >> aSignal) numArgs < argPattern max])
			ifTrue: [Error signal: 'Could not connect sender to receiver!! Pattern is invalid!!']].

	((self signalConnections
		at: aSignal
		ifAbsentPut: [WeakIdentityKeyDictionary new])
		at: aReceiver
		ifAbsentPut: [OrderedCollection new])
		add: aSelector -> argPattern.! !

!Object methodsFor: '*signals-core' stamp: 'mt 6/2/2010 22:41'!
disconnectAllSignals
	
	SignalConnectionsRepository instance
		removeKey: self
		ifAbsent: [^ false].
		
	^ true! !

!Object methodsFor: '*signals-core' stamp: 'mt 6/2/2010 21:35'!
disconnectSignal: aSignal
	
	self signalConnections
		removeKey: aSignal
		ifAbsent: [^ false].
		
	^ true! !

!Object methodsFor: '*signals-core' stamp: 'mt 6/2/2010 21:35'!
disconnectSignal: aSignal from: aReceiver
	
	(self signalConnections includesKey: aSignal)
		ifFalse: [^ false].
		
	(self signalConnections at: aSignal)
		removeKey: aReceiver
		ifAbsent: [^ false].
	
	(self signalConnections at: aSignal)
		ifEmpty: [^ self disconnectSignal: aSignal].
		
	^ true! !

!Object methodsFor: '*signals-core' stamp: 'mt 6/2/2010 22:13'!
disconnectSignal: aSignal from: aReceiver selector: aSelector
	
	(self signalConnections includesKey: aSignal)
		ifFalse: [^ false].
		
	((self signalConnections at: aSignal)
		at: aReceiver
		ifAbsent: [^ false]) "weak"
		removeAllSuchThat: [:selectorAndPattern | selectorAndPattern key = aSelector].
	
	((self signalConnections at: aSignal)
		at: aReceiver
		ifAbsent: [^ true]) "weak"
		ifEmpty: [self disconnectSignal: aSignal from: aReceiver].
	
	^ true! !

!Object methodsFor: '*signals-core' stamp: 'mt 6/2/2010 22:18'!
emit
	"Should be called from signals only."

	| signalCtx signal args |
	signalCtx := thisContext sender.
	signal := signalCtx method.
	
	"Extract arguments."
	args := Array new: signal numArgs.
	(1 to: signal numArgs) do: [:index |
		args at: index put: (signalCtx tempAt: index)].
	
	"Fetch targets and dispatch."
	(self signalConnections at: signal selector ifAbsent: [^ self]) associationsDo: [ :association |
		association key ifNotNil: [ :receiver | "Don't let the key go away -> weak dictionary"
			association value do: [ :selectorWithPattern |
				| selector pattern filteredArgs |
				selector := selectorWithPattern key.
				pattern := selectorWithPattern value.
				pattern
					ifNil: [receiver perform: selector withArguments: args]
					ifNotNil: [
						filteredArgs := Array new: pattern size.
						pattern withIndexDo: [:source :target | filteredArgs at: target put: (args at: source)].
						receiver perform: selector withArguments: filteredArgs] ] ] ].
	! !

!Object methodsFor: '*signals-core-private' stamp: 'mt 6/2/2010 22:41'!
signalConnections

	^ SignalConnectionsRepository instance
		at: self
		ifAbsentPut: [Dictionary new]! !

Object subclass: #SignalConsumer
	instanceVariableNames: 'value switch'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Signals-Tests'!

!SignalConsumer methodsFor: 'slots' stamp: 'mt 6/2/2010 21:43'!
switchState

	^ switch ifNil: [switch := #off]! !

!SignalConsumer methodsFor: 'slots' stamp: 'mt 6/2/2010 21:44'!
toggleSwitch

	switch := self switchState = #off
		ifTrue: [#on]
		ifFalse: [#off].! !

!SignalConsumer methodsFor: 'slots' stamp: 'mt 6/2/2010 21:43'!
value

	^ value! !

!SignalConsumer methodsFor: 'slots' stamp: 'mt 6/2/2010 21:43'!
value: aValue

	value := aValue.! !

Object subclass: #SignalProducer
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Signals-Tests'!

!SignalProducer methodsFor: 'signals' stamp: 'mt 6/2/2010 21:39'!
newSignalProduced

	self emit.! !

!SignalProducer methodsFor: 'signals' stamp: 'mt 6/2/2010 21:39'!
valueChanged: aValue

	self emit.! !

!SignalProducer methodsFor: 'signals' stamp: 'mt 6/2/2010 22:20'!
valueChanged: aValue and: anotherValue

	self emit.! !

WeakIdentityKeyDictionary subclass: #SignalConnectionsRepository
	instanceVariableNames: ''
	classVariableNames: 'SignalConnectionsRepository'
	poolDictionaries: ''
	category: 'Signals-Core'!

!SignalConnectionsRepository commentStamp: 'mt 6/2/2010 22:43' prior: 0!
I store all active signal-selector connections with weak data structures.!

!SignalConnectionsRepository class methodsFor: 'instance' stamp: 'mt 6/2/2010 21:28'!
instance

	^ SignalConnectionsRepository ifNil: [SignalConnectionsRepository := self new]! !
